#include "field.h"
#include "support.h"
#include "images.h"
#include <math.h>

#include "settings.h"

int itemsize = 30;

/*
 * If abs(deadmatch) is bigger then some constant
 * on of the player gets the penalty. It depends on the sign
 * of deadmatch.
 */
int deadmatch = 0;

extern GtkWidget*app;

char*getcolorname(int color);
void set_text(GtkLabel*l,int x);
void set_text2(GtkLabel*l,int m,int x);

void set_text(GtkLabel*l,int x)
{
        gchar s[20];
        g_snprintf(s,20,"%d",x);
        gtk_label_set_text(l,s);
}
void set_text2(GtkLabel*l,int m,int x)
{
        gchar s[40];
        g_snprintf(s,40,"%d x %d",m,x);
        gtk_label_set_text(l,s);
}

Field*field_new(GnomeCanvas*c,GnomeCanvas*preview,GtkLabel*j,GtkLabel*p,GtkLabel*t,GtkLabel*l)
{
        Field*f = g_new(Field,1);
        f->canvas = c;
        f->jewelsview = j;
        f->pointsview = p;
        f->thisptsview = t;
        f->levelview = l;
        memset(&(f->floater),0,sizeof(GewelsFloater));
        f->floater.canvas = preview;
        f->deadmatchsign = 0;
        f->state = gameover;
        f->root = NULL;
        f->pointsflash = NULL;
        f->pointsflashlist = NULL;
        return f;
}

void field_init(Field*f,int deadmtchsign,int scorelimit,int computer)
{
        int i,j;
        
        if(NULL!=f->root){
                gtk_object_destroy(GTK_OBJECT(f->root));
                f->root = NULL;
        }
        for(j = 0;j<SIZE_Y;j++)
                for(i = 0;i<SIZE_X;i++)
                        f->field[i][j].state = empty;
        f->sub = 0;
        f->root = new_background(gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                                 SIZE_X*itemsize-1,(SIZE_Y-3)*itemsize-1);
        f->state = fall;
        for(i = 0;i<3;i++){
                if(NULL!=f->floater.preview[i]){
                        gtk_object_destroy(GTK_OBJECT(f->floater.preview[i]));
                        f->floater.preview[i] = NULL;
                }
                f->floater.next[i] = rand()%COLOR_COUNT;
        }
        f->fastpoints = f->points = f->jewels = 0;
        f->mul = 1;
        set_text(f->levelview,f->level = prefs.startinglevel);
        deadmatch = 0;
        set_faces(deadmatch,f->deadmatchsign);
        f->deadmatchsign = deadmtchsign;
        f->scorelimit = scorelimit;
        f->tvuse = 0;
        f->computer = computer;
        gettimeofday(&(f->last),NULL);
        f->pointsflash = NULL;
        f->pointsflashlist = NULL;
}

void field_destroy(Field*f)
{
        if(NULL!=f->root){
                gtk_object_destroy(GTK_OBJECT(f->root));
                f->root = NULL;
        }
        g_free(f);
}


//fixme there must be an array and I must use %
char*getcolorname(int color)
{
        switch(color){
        case 0:   return "cyan";
        case 1:   return "blue";
        case 2:   return "green";
        case 3:   return "yellow";
        case 4:   return "red";
        case 5:   return "magenta";
        }                
        return "cyan";
}

GnomeCanvasItem*field_add(Field*f,int x,int y,int color)
{
        if(empty!=f->field[x][y].state)
                return NULL;
        f->field[x][y].state = falling;
        f->field[x][y].color = color;
        f->field[x][y].item = paint_item(
                gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                x,y,color,0);
        return f->field[x][y].item;
}

static void move_down(Field*f,int x,int y);

void move_down(Field*f,int x,int y)
{
        if(laying==f->field[x][y+1].state){
                f->field[x][y].state = laying;
                gnome_canvas_item_move(f->field[x][y].item,0,f->smallstep+itemsize-f->sub);
        } else {
                // there must be an empty slot...
                // comment it out!
                if(empty!=f->field[x][y+1].state)
                        g_print("Error in field.c::down\n");
                f->field[x][y+1] = f->field[x][y];
                gnome_canvas_item_move(f->field[x][y].item,0,f->smallstep);
                f->field[x][y].state = empty;
                f->field[x][y].item = NULL;
        }
}

int field_fall(Field*f);

int field_fall(Field*f)
{
        int falling_counter = 0;
        int i,j;
        
        f->sub+= f->smallstep;
        if(f->sub>=itemsize){
                // attention: I have named outer cycle variable j and inner i
                for(i = 0;i<SIZE_X;i++)
                        if(falling==f->field[i][SIZE_Y-1].state){
                                falling_counter++;
                                f->field[i][SIZE_Y-1].state = laying;
                                gnome_canvas_item_move(f->field[i][SIZE_Y-1].item,
                                                       0,f->smallstep+itemsize-f->sub);
                        }
                for(j = SIZE_Y-2;j>=0;j--)
                        for(i = 0;i<SIZE_X;i++)
                                if(falling==f->field[i][j].state){
                                        falling_counter++;
                                        move_down(f,i,j);
                                }
                f->sub-= itemsize;
        }
        else{
                for(j = 0;j<SIZE_Y;j++)
                        for(i = 0;i<SIZE_X;i++)
                                if(falling==f->field[i][j].state){
                                        falling_counter++;
                                        gnome_canvas_item_move(f->field[i][j].item,
                                                               0,f->smallstep);
                                }                
        }        
        return falling_counter ? 0 : 1;
}
int field_mark_helper(Field*f);

int field_mark(Field*f);

int field_mark_helper(Field*f)
{
        int i,j;
        int last;
        int count;
        int pts = 0;
/*
        for(j = 0;j<SIZE_Y;j++){
                for(i = 0;i<SIZE_X;i++){
                        switch(f->field[i][j].state){
                        case laying:
                                g_print("%d ",f->field[i][j].color);
                                break;
                        case empty:
                                g_print(". ");
                                break;
                        case falling:
                                g_print("_ ");
                                break;
                        case marked:
                                g_print("x ");
                                break;
                        }
                }
                g_print("\n");
        }
*/  

#define REMOVE  if(count>=3){ pts+= count;\
                for(;count>0;count--){\
                        f->field[i-count][j].state = marked; \
                        }}
        
        // horizontal...
        for(j = 0;j<SIZE_Y;j++){
                count = 0;
                last = -1;
                for(i = 0;i<SIZE_X;i++){
                        if(laying==f->field[i][j].state || marked==f->field[i][j].state){
                                if(last==f->field[i][j].color)
                                        count++;
                                else{
                                        REMOVE;
                                        last = f->field[i][j].color;
                                        count = 1;
                                }
                        } else {
                                REMOVE;
                                count = 0;
                                last = -1;
                        }
                }
                REMOVE;
        }
#undef  REMOVE
#define REMOVE  if(count>=3){ pts+= count;\
                for(;count>0;count--){\
                        f->field[i][j-count].state = marked; \
                        }}
        
        // vertical...
        for(i = 0;i<SIZE_X;i++){
                count = 0;
                last = -1;
                for(j = 0;j<SIZE_Y;j++){
                        if(laying==f->field[i][j].state || marked==f->field[i][j].state){
                                if(last==f->field[i][j].color)
                                        count++;
                                else{
                                        REMOVE;
                                        last = f->field[i][j].color;
                                        count = 1;
                                }
                        } else {
                                REMOVE;
                                count = 0;
                                last = -1;
                        }
                }
                REMOVE;
        }
#undef  REMOVE

#define REMOVE  if(count>=3){ pts+= count;\
                for(;count>0;count--){\
                        f->field[i-count][j+i-count].state = marked; \
                        }}
        
        // diagonal '\'
        for(j = -SIZE_X;j<SIZE_Y;j++){
                count = 0;
                last = -1;
                for(i = 0;i<SIZE_X;i++){
                        if(i+j>=0 && i+j<SIZE_Y){
                                if(laying==f->field[i][i+j].state ||
                                   marked==f->field[i][i+j].state){
                                        if(last==f->field[i][i+j].color)
                                                count++;
                                        else{
                                                REMOVE;
                                                last = f->field[i][i+j].color;
                                                count = 1;
                                        }
                                } else {
                                        REMOVE;
                                        count = 0;
                                        last = -1;
                                }
                        } else {
                                REMOVE;
                        }
                }
                REMOVE;
        }
#undef  REMOVE
#define REMOVE  if(count>=3){ pts+= count;\
                for(;count>0;count--){\
                        f->field[i-count][j-i+count].state = marked; \
                        }}
        
        // diagonal '/'
        for(j = 0;j<SIZE_Y+SIZE_X;j++){
                count = 0;
                last = -1;
                for(i = 0;i<SIZE_X;i++){
                        if(j-i>=0 && j-i<SIZE_Y){
                                if(laying==f->field[i][j-i].state ||
                                   marked==f->field[i][j-i].state){
                                        if(last==f->field[i][j-i].color)
                                                count++;
                                        else{
                                                REMOVE;
                                                last = f->field[i][j-i].color;
                                                count = 1;
                                        }
                                } else {
                                        REMOVE;
                                        count = 0;
                                        last = -1;
                                }
                        } else {
                                REMOVE;
                        }
                }
                REMOVE;
        }
#undef  REMOVE

        return pts;
}

void field_flashpts(Field*f,int pts)
{
        if(NULL!=f->pointsflash){
                gtk_object_destroy(GTK_OBJECT(f->pointsflash));
                while(NULL!=f->pointsflashlist){
                        FlashPointItem*d = f->pointsflashlist;
                        f->pointsflashlist = f->pointsflashlist->next;
                        free(d);
                }
        }

        f->pointsflash = gnome_canvas_item_new(
                gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                gnome_canvas_group_get_type(),
                "x",(double)0,//(SIZE_X*itemsize/2),
                "y",(double)((SIZE_Y-7)*itemsize/2),
                NULL);
        f->pointsflashlist = new_score_flash((GnomeCanvasGroup*)f->pointsflash,pts);
}

void field_flashpts_remove(Field*f)
{
        if(NULL!=f->pointsflash){
                gtk_object_destroy(GTK_OBJECT(f->pointsflash));
                while(NULL!=f->pointsflashlist){
                        FlashPointItem*d = f->pointsflashlist;
                        f->pointsflashlist = f->pointsflashlist->next;
                        free(d);
                }
                f->pointsflash = NULL;
                f->pointsflashlist = NULL;
        }
}
void field_flashpts_animate(Field*f)
{
        if(NULL!=f->pointsflashlist){
                FlashPointItem*walk = f->pointsflashlist;
                int y = 0;
                while(NULL!=walk){
                        y = sin((double)(walk->x+= 5)/(double)(itemsize))*
                                ((double)(itemsize << 1));
                        gnome_canvas_item_move(walk->item,5,walk->y-y);
                        walk->y = y;
                        walk = walk->next;
                }
        }
}

int field_mark(Field*f)
{
        int pts = field_mark_helper(f);
        
        f->disappear = 1;
        f->subdisappear = 5;

        if(pts){
                pts*= 100;
                pts+= (f->fastpoints << 2)/itemsize;
                field_flashpts (f, pts*f->mul);
                f->points+= pts*f->mul;
                f->fastpoints = 0;
                set_text(f->pointsview,f->points);
                set_text2(f->thisptsview,f->mul,pts);
                return 1;
        }
        
        return 0;
}
int field_disappear(Field*f);

int field_disappear(Field*f)
{
        int i,j;
        int mark;

        if(!--f->subdisappear){
                f->disappear++;
                f->subdisappear = 3;
                if(5>=f->disappear){
                        for(i = 0;i<SIZE_X;i++){
                                for(j = SIZE_Y-1;j>=0;j--){
                                        if(marked==f->field[i][j].state){
                                                gtk_object_destroy(GTK_OBJECT(
                                                        f->field[i][j].item));
                                                f->field[i][j].item = paint_item(
                                                        gnome_canvas_root(
                                                                GNOME_CANVAS(f->canvas)),
                                                        i,j+1,f->field[i][j].color,f->disappear);
                                        }
                                }
                        }       
                } else {
                        int jws = 0;
                        for(i = 0;i<SIZE_X;i++){
                                mark = 0;
                                for(j = SIZE_Y-1;j>=0;j--){
                                        if(marked==f->field[i][j].state){
                                                jws++;
                                                mark = 1;
                                                gtk_object_destroy(
                                                        GTK_OBJECT(f->field[i][j].item));
                                                f->field[i][j].state = empty;
                                        }
                                        else if(mark && laying==f->field[i][j].state)
                                                f->field[i][j].state = falling;
                                }
                        }
                        if(jws){
                                set_text(f->jewelsview,f->jewels+= jws);
                                deadmatch+= f->deadmatchsign*jws;
                                set_faces(deadmatch,f->deadmatchsign);
                        }
                        
                return 1;
                }
                
        }
        return 0;
}
void field_add_penalty(Field*f)
{
        int i;
        int r;
        for(i = 0;i<SIZE_X;i++){
                if(i==f->floater.x) // do not insert penalty into the column floater is in
                        continue;
                r = rand() % (COLOR_COUNT << 1);
                if(r<COLOR_COUNT){
                        f->field[i][0].color = r;
                        f->field[i][0].state = falling;
                        f->field[i][0].item = paint_item(
                                gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                                i,1,f->field[i][0].color,0);
                }
        }
}

int field_step(Field*f)
{
        switch(f->state) {
        case floater:
                if(f->deadmatchsign*deadmatch<0 &&
                   (deadmatch>0 ? deadmatch : -deadmatch)>DEADMATCH) {
                        deadmatch+= DEADMATCH*f->deadmatchsign;
                        f->state = deadmatchdisappear;
                        field_add_penalty(f);
                        set_faces(deadmatch,f->deadmatchsign);
                }else if(gewels_floater_step(f)){
                        f->state = mark;
                        gnome_triggers_do(NULL,NULL,"gewels","place",NULL);
                }
                break;
        case mark:
                if(field_mark(f)){
                        f->state = disappear;
                        gnome_triggers_do(NULL,NULL,"gewels","kill",NULL);
                }else{
                        if(gewels_floater_add(f))
                                f->state = gameover;
                        else{
                                f->mul = 1;
                                f->sub = 0;
                                f->state = floater;
                                if(f->jewels>=JEWELS_IN_ONE_LEVEL*(f->level-prefs.startinglevel+1)){
                                        if(f->level<LEVEL_COUNT)
                                                f->level++;
                                        set_text(f->levelview,f->level);
                                }
                        }
                }
                break;
        case disappear:
                field_flashpts_animate(f);
                if(field_disappear(f)){
                        f->state = fall;
                        f->sub = itemsize;
                }
                break;
        case fall:
                field_flashpts_animate(f);
                if(field_fall(f)){
                        f->mul<<= 1;
                        f->state = mark;
                        field_flashpts_remove(f);
                }
                break;
        case deadmatchdisappear:
                if(field_disappear(f)){
                        f->state = deadmatchfall;
                        f->subcache = f->sub;
                        f->sub = itemsize;
                }
                break;
        case deadmatchfall:
                if(field_fall(f)){
                        f->mul<<= 1;
                        f->sub = f->subcache;
                        f->state = floater;
                        if(f->computer)
                                solve(f);
                }
                break;
        case gameover:
                return 1;
        }
        return 0;
}
int gewels_floater_add(Field*f)
{
        GewelsFloater*fl = &(f->floater);
        int i;
        
        if(empty!=f->field[3][1].state ||
           empty!=f->field[3][2].state ||
           empty!=f->field[3][3].state)
                return 1;

        fl->x = 3;
        fl->y[0] = 1;
        fl->y[1] = 2;
        fl->y[2] = 3;

        fl->f[0].state = falling;
        fl->f[0].color = fl->next[0];
        fl->f[0].item = paint_item(
                gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                3,fl->y[0],fl->f[0].color,0);
        fl->f[1].state = falling;
        fl->f[1].color = fl->next[1];
        fl->f[1].item = paint_item(
                gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                3,fl->y[1],fl->f[1].color,0);
        fl->f[2].state = falling;
        fl->f[2].color = fl->next[2];
        fl->f[2].item = paint_item(
                gnome_canvas_root(GNOME_CANVAS(f->canvas)),
                3,fl->y[2],fl->f[2].color,0);

        for(i = 0;i<3;i++){
                if(NULL!=fl->preview[i])
                        gtk_object_destroy(GTK_OBJECT(fl->preview[i]));
                fl->preview[i] = paint_item(
                        gnome_canvas_root(fl->canvas),
                        0,i+4,fl->next[i] = rand()%COLOR_COUNT,0);
        }

        f->distance = 0;
        gettimeofday(&f->first,NULL);

        if(f->computer)
                solve(f);
        
        return 0;
}
int gewels_floater_down(Field*f)
{
        GewelsFloater*fl = &(f->floater);
        
        if(fl->y[2]>=SIZE_Y-1 || laying==f->field[fl->x][fl->y[2]+1].state){
                fl->f[0].state = fl->f[1].state = fl->f[2].state = laying;
                f->field[fl->x][fl->y[0]] = fl->f[0];
                f->field[fl->x][fl->y[1]] = fl->f[1];
                f->field[fl->x][fl->y[2]] = fl->f[2];
                return 1;
        }
        fl->y[0]++;
        fl->y[1]++;
        fl->y[2]++;        

        return 0;
}
int gewels_floater_step(Field*f)
{
        GewelsFloater*fl = &(f->floater);
        int i;
        int ret = 0;
        
        f->sub+= f->smallstep;
        f->distance+= f->smallstep;
        if(f->sub>=itemsize){
                int step = (ret = gewels_floater_down(f)) ?
                        f->smallstep+itemsize-f->sub : f->smallstep;
                for(i = 0;i<3;i++)
                        gnome_canvas_item_move(fl->f[i].item,0,step);
                f->sub-= itemsize;
        }
        else{
                for(i = 0;i<3;i++)
                        gnome_canvas_item_move(fl->f[i].item,0,f->smallstep);
        }
        return ret;
}

void field_move_floater(Field*f,UserInput in)
{
        if(floater!=f->state)
                return ;
        gewels_floater_move(f,in);
}

void gewels_floater_move(Field*f,UserInput in)
{
        GewelsFloater*fl = &(f->floater);
        FieldItem x;
        
        gnome_triggers_do(NULL,NULL,"gewels","move",NULL);
        switch(in){
        case up:
                gnome_canvas_item_move(fl->f[0].item,0,itemsize);
                gnome_canvas_item_move(fl->f[1].item,0,itemsize);
                gnome_canvas_item_move(fl->f[2].item,0,-itemsize << 1);
                x = fl->f[2];
                fl->f[2] = fl->f[1];
                fl->f[1] = fl->f[0];
                fl->f[0] = x;
                break;
        case down:
                gnome_canvas_item_move(fl->f[0].item,0,itemsize << 1);
                gnome_canvas_item_move(fl->f[1].item,0,-itemsize);
                gnome_canvas_item_move(fl->f[2].item,0,-itemsize);
                x = fl->f[0];
                fl->f[0] = fl->f[1];
                fl->f[1] = fl->f[2];
                fl->f[2] = x;
                break;
        case left:
                if(fl->x<=0)
                        break;
                if(empty!=f->field[fl->x-1][fl->y[2]].state)
                        break;
                fl->x--;
                gnome_canvas_item_move(fl->f[0].item,-itemsize,0);
                gnome_canvas_item_move(fl->f[1].item,-itemsize,0);
                gnome_canvas_item_move(fl->f[2].item,-itemsize,0);
                break;
        case right:
                if(fl->x+1>=SIZE_X)
                        break;
                if(empty!=f->field[fl->x+1][fl->y[2]].state)
                        break;
                fl->x++;
                gnome_canvas_item_move(fl->f[0].item,itemsize,0);
                gnome_canvas_item_move(fl->f[1].item,itemsize,0);
                gnome_canvas_item_move(fl->f[2].item,itemsize,0);
                break;
        case hint:
                solve(f);
                break;
        }
}
/*
 * helper function for solve
 */
int get_length(Field*f,int column,int floater,int y);

int get_length(Field*f,int column,int floater,int y)
{
        Field tmp;

        if(y<2)
                return 0;
        
        memcpy(tmp.field,f->field,sizeof f->field);
        switch(floater){
        case 0:
                tmp.field[column][y] = f->floater.f[2];
                tmp.field[column][y-1] = f->floater.f[1];
                tmp.field[column][y-2] = f->floater.f[0];
                break;
        case 1:
                tmp.field[column][y] = f->floater.f[1];
                tmp.field[column][y-1] = f->floater.f[0];
                tmp.field[column][y-2] = f->floater.f[2];                
                break;
        case 2:
                tmp.field[column][y] = f->floater.f[0];
                tmp.field[column][y-1] = f->floater.f[2];
                tmp.field[column][y-2] = f->floater.f[1];                
                break;                
        }
        tmp.field[column][y].state =
                tmp.field[column][y-1].state =
                tmp.field[column][y-2].state = 
                laying;
        
        return field_mark_helper(&tmp);
}
/*
 * find the best position to place floater to
 */
void solve(Field*f)
{
        int y[SIZE_X]; // y index to field array to place floater to.
        int results[SIZE_X][3];  // how many items disappear if I place floater here...
        int i,j;
        int besti,bestj,bestlen,besty;
        int fi[] = { 0, 5, 1, 4, 2, 3 
        };
        
         
        for(i = 0;i<SIZE_X;i++){
                for(j = 0;j<SIZE_Y;j++)
                        if(empty!=f->field[i][j].state){
                                y[i] = j-1;
                                break;
                        }
                if(SIZE_Y==j)
                        y[i] = j-1;
        }
        // find the best place to place floater
        for(i = 0;i<SIZE_X;i++)
                for(j = 0;j<3;j++)
                        results[i][j] = get_length(f,i,j,y[i]);
        

        besti = bestj = bestlen = besty = 0;
        for(j = 0;j<3;j++){
                for(i = 0;i<SIZE_X;i++){
                        if(results[fi[i]][j]>bestlen){
                                bestlen = results[fi[i]][j];
                                besti = fi[i];
                                bestj = j;
                                besty = y[fi[i]];
                        }else if(results[fi[i]][j]==bestlen && besty<y[fi[i]]){
                                bestlen = results[fi[i]][j];
                                besti = fi[i];
                                bestj = j;
                                besty = y[fi[i]];
                        }
                }
        }

        if(!bestlen){
                // if 2 colors are same, we want them up
                if(f->floater.f[1].color==f->floater.f[2].color)
                        bestj = 2;
                else if(f->floater.f[0].color==f->floater.f[2].color)
                        bestj = 1;
                else{
                        /* try to find this:
                         *
                         *     x<-+
                         *     x<-|
                         *     o<--- floater
                         *   x x<---  field
                         *   x o<-+
                         */
                        if(besty<SIZE_Y-1){
                                int c = f->field[besti][besty].color;
                                if(c==f->floater.f[0].color)
                                        bestj = 2;
                                else if(c==f->floater.f[2].color)
                                        bestj = 1;
                                else if(besty<SIZE_Y-2){
                                        c = f->field[besti][besty+1].color;
                                        if(c==f->floater.f[0].color)
                                                bestj = 2;
                                        else if(c==f->floater.f[1].color)
                                                bestj = 1;
                                }
                        }
                }
        }

        while(bestj--)
                gewels_floater_move(f,up);

        besty = f->floater.x-besti;
        if(besty>0)
                while(besty-->0)
                        gewels_floater_move(f,left);
        else
                while(besty++<0)
                        gewels_floater_move(f,right);
}

int field_computer_wants_fast(Field*f)
{
        if(!f->computer)
                return 0;
        
        if(rand()%200<prefs.computerskills)
                return 1;
        
        return 0;
}

