#ifndef __FIELD_H__
#define __FIELD_H__

#include <gnome.h>
#include <sys/time.h>
#include "settings.h"
#include "images.h"

extern int itemsize;

typedef enum
{
        empty = 0,
        falling,
        laying,
        marked,
}FieldItemState;

typedef struct
{
        FieldItemState state;
        
        int color;// index, type or whatever you want to call it
        GnomeCanvasItem*item; // corresponding canvas item...
}FieldItem;

typedef struct
{
        FieldItem f[3];
        
        int x;
        int y[3];

        int next[3];
        GnomeCanvasItem*preview[3];
        GnomeCanvas*canvas;
}GewelsFloater;

typedef enum
{
        floater,
        mark,
        disappear,
        fall,
        gameover,
        deadmatchdisappear,
        deadmatchfall,
}FieldState;

typedef struct
{
        GnomeCanvas*canvas;
        GnomeCanvasItem*root;
        GnomeCanvasItem*pointsflash;
        FlashPointItem*pointsflashlist;

        FieldState state;
        
        FieldItem field[SIZE_X][SIZE_Y];
        int sub;
        int disappear;
        int subdisappear;

        GewelsFloater floater;

        int points;
        int fastpoints;
        int jewels;
        int mul;
        int level;

        int scorelimit;
        
        int fast;
        int smallstep;

        int subcache;

        // In fact this is the position of player. Left is -1 and right is 1
        int deadmatchsign;

        // score, ...
        GtkLabel*jewelsview;
        GtkLabel*pointsview;
        GtkLabel*levelview;
        GtkLabel*thisptsview;

        struct timeval first; // when the floater was added
        int distance; // the distance floater already went
        
        struct timeval last;
        int tvuse;
        struct timeval tv;

        int computer;  // level of computer skills. 0 means computer does not play.
}Field;

typedef enum
{
        left,
        right,
        up,
        down,
        hint,
}UserInput;

Field*field_new(GnomeCanvas*c,GnomeCanvas*preview,GtkLabel*j,GtkLabel*p,GtkLabel*t,GtkLabel*l);
void field_init(Field*f,int deadmtchsign,int scorelimit,int computer);
void field_destroy(Field*f);
void field_move_floater(Field*f,UserInput in);
void field_flash_pts(Field*f,int pts);
void field_flashpts_remove(Field*f);
void field_flashpts_animate(Field*f);

GnomeCanvasItem*field_add(Field*f,int x,int y,int color);
int field_step(Field*f);

int field_computer_wants_fast(Field*f);

void field_add_penalty(Field*f);

int gewels_floater_add(Field*f);
int gewels_floater_down(Field*f);
void gewels_floater_move(Field*f,UserInput in);
int gewels_floater_step(Field*f);

void solve(Field*f);

#endif//__FIELD_H__
