#include "game.h"
#include "field.h"

#include "support.h"
#include "images.h"

Field*fieldleft = NULL;
Field*fieldright = NULL;

static gint game_step_new(gpointer data);

/*
 * send user input to the field
 */
void game_move(int game,UserInput in)
{
        if(!game)
                field_move_floater(fieldleft,in);
        else if(1==game)
                field_move_floater(fieldright,in);
}
/* 
 * set fast movement in field
 */
void game_set_fast(int game )
{
        if(!game){
                fieldleft->fast = 1;
        }
        else if(1==game){
                fieldright->fast = 1;
        }
}
/*
 * make one step :-)
 */
static gint game_step_new(gpointer data)
{
        Field*f = (Field*)data;
        struct timeval now;
        int speed;
        
        gettimeofday(&now,NULL);

        switch(f->state){
        case floater:
                if(f->scorelimit && f->jewels>=f->scorelimit){
                        int pos;
                        f->state = gameover;
                        pos = gnome_score_log(f->points,NULL,TRUE);
                }
                
                if(!f->tvuse){
                        // measurement of speed
                        f->tv = now;
                }
                f->tvuse++;

                if(f->fast){
                        speed = 1;
                        f->smallstep = 3;
                        f->fastpoints++;
                }else{
                        // distance from the point of insertion
                        int dist;
                        double t = (double)((now.tv_usec-f->first.tv_usec)/1000.0);
                        t+= 1000*(now.tv_sec-f->first.tv_sec);
                        dist = itemsize*(0.5+f->level/20.0)*(t/1000);
                        // now t is time since floater was inserted into field
                        f->smallstep = dist-f->distance;
                        speed = 30;

                        if(field_computer_wants_fast(f))
                                f->fast = 1;
                }
                break;
        case gameover:
                gnome_score_log(f->points,"1",TRUE);
                if(fieldleft==f)
                        gnome_app_flash((GnomeApp*)app,_("Left player: game over."));
                else
                        gnome_app_flash((GnomeApp*)app,_("Right player: game over."));
                
                return 0;
        case deadmatchdisappear:
        case deadmatchfall:
                speed = 5;
                f->smallstep = 15;
                break;
        default:
                if(f->tvuse){
                        struct timeval t2 = now;
                        t2.tv_sec-= f->tv.tv_sec;
                        if((t2.tv_usec-= f->tv.tv_usec)<0){
                                t2.tv_sec--;
                                t2.tv_usec+= 1000000;
                        }
//                        g_print("Level %02d: %02ld.%06ld\t%d\n",
//                                f->level,t2.tv_sec,t2.tv_usec,f->tvuse);
                        f->tvuse = 0;
                }
                speed = 10;
                f->smallstep = 5;
                f->fast = 0;
        }

        f->last = now;
        gtk_timeout_add(speed,game_step_new,f);
        field_step(f);
        
        return 0;
}
/*
 * Create new instance of field
 */
void game_new()
{
        fieldleft = field_new((GnomeCanvas*)lookup_widget(app,"field"),
                              (GnomeCanvas*)lookup_widget(app,"preview"),
                              (GtkLabel*)lookup_widget(app,"jewels"),
                              (GtkLabel*)lookup_widget(app,"points"),
                              (GtkLabel*)lookup_widget(app,"thispts"),
                              (GtkLabel*)lookup_widget(app,"level")
                );
        fieldright = field_new((GnomeCanvas*)lookup_widget(app,"field2"),
                               (GnomeCanvas*)lookup_widget(app,"preview2"),
                               (GtkLabel*)lookup_widget(app,"jewels2"),
                               (GtkLabel*)lookup_widget(app,"points2"),
                               (GtkLabel*)lookup_widget(app,"thispts2"),
                               (GtkLabel*)lookup_widget(app,"level2")
                );
}

/*
 * destroy field
 */
void game_destroy()
{
        g_free(fieldleft);
        g_free(fieldright);
}

void reload_images_if_possible(void );

void reload_images_if_possible()
{
        if(gameover==fieldleft->state && gameover==fieldright->state)
                reload_images();
}

/*
 * start a new game
 */
void game_new_left(void )
{
        reload_images_if_possible();
        
        if(gameover!=fieldleft->state)
                return ;
        field_init(fieldleft,0,0,0);

        game_step_new(fieldleft);
        gnome_app_flash((GnomeApp*)app,_("Game started in the left field."));
}

void game_new_right(void )
{
        reload_images_if_possible();

        if(gameover!=fieldright->state)
                return ;
        field_init(fieldright,0,0,0);

        game_step_new(fieldright);
        gnome_app_flash((GnomeApp*)app,_("Game started in the right field."));
}
void game_new_deadmatch(void )
{
        reload_images_if_possible();

        if(gameover!=fieldleft->state && gameover!=fieldright->state)
                return ;
        field_init(fieldleft,-1,0,0);
        field_init(fieldright,1,0,0);

        game_step_new(fieldleft);
        game_step_new(fieldright);
        gnome_app_flash((GnomeApp*)app,_("Deadmatch game started."));
}
/*
 * start a new 200-contest game in right field
 */
void game_new_200_right(void )
{
        reload_images_if_possible();

        if(gameover!=fieldright->state)
                return ;
        field_init(fieldright,0,200,0);

        game_step_new(fieldright);
        gnome_app_flash((GnomeApp*)app,_("200 contest started in the right field."));        
}
/*
 * start a new 200-contest game in left field
 */
void game_new_200_left(void )
{
        reload_images_if_possible();

        if(gameover!=fieldleft->state)
                return ;
        field_init(fieldleft,0,200,0);

        game_step_new(fieldleft);
        gnome_app_flash((GnomeApp*)app,_("200 contest started in the left field."));
}

void game_new_human_vs_comp(void )
{
        reload_images_if_possible();

        if(gameover!=fieldleft->state && gameover!=fieldright->state)
                return ;
        field_init(fieldleft,-1,0,0);
        field_init(fieldright,1,0,1);

        game_step_new(fieldleft);
        game_step_new(fieldright);
        gnome_app_flash((GnomeApp*)app,_("Deadmatch game Human vs. Computer started."));
}
void game_new_comp_vs_human(void )
{
        reload_images_if_possible();

        if(gameover!=fieldleft->state && gameover!=fieldright->state)
                return ;
        field_init(fieldleft,-1,0,1);
        field_init(fieldright,1,0,0);

        game_step_new(fieldleft);
        game_step_new(fieldright);
        gnome_app_flash((GnomeApp*)app,_("Deadmatch game Computer vs. Human started."));
}
void game_new_comp_vs_comp(void )
{
        reload_images_if_possible();

        if(gameover!=fieldleft->state && gameover!=fieldright->state)
                return ;
        field_init(fieldleft,-1,0,1);
        field_init(fieldright,1,0,1);

        game_step_new(fieldleft);
        game_step_new(fieldright);
        gnome_app_flash((GnomeApp*)app,_("Deadmatch game Computer vs. Computer started."));
}
